import inspect
from dataclasses import dataclass

from with_partial.partial_proxy import PartialProxy
import typing
from functools import partial
import addict


@dataclass()
class _CallPartialProxy(PartialProxy):
    """
    A slightly modified proxy that, when called() will forward its arguments
    and itself to a hook function
    """

    call_hook: typing.Callable = lambda *args, **kwargs: None

    def __call__(self, *args, **kwargs):
        return self.call_hook(self, *args, **kwargs)


class PartialContext:
    """
    A context manager class for accessing functions as partials
    """
    def __init__(self, globs: dict = None, locs: dict = None):
        # Store the locals and globals dictionaries if provided
        if globs is not None or locs is not None:
            self.frame = addict.Dict()
        else:
            self.frame = None
        if globs is not None:
            self.frame.update(globs)
        if locs is not None:
            self.frame.update(locs)

    def __enter__(self):
        # We first get the frame local variables. If they were provided at
        # instantiation time, use those. Otherwise get them using inspect
        if self.frame is None:
            stack = inspect.stack(context=0)
            locs = addict.Dict(__builtins__, stack[1].frame.f_locals)
            del stack
        else:
            locs = self.frame

        # We define our hook function that is called when the user calls()
        # the context manager
        def eval(proxy: PartialProxy, *args, **kwargs) -> typing.Callable:
            # Find the actual function they are trying to call
            func = proxy.eval(locs)
            # Return a partially applied version
            return partial(func, *args, **kwargs)

        # Finally, we return a proxy for the user to access
        return _CallPartialProxy(path=[], call_hook=eval)

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass


import operator
import typing
from dataclasses import dataclass, field


@dataclass
class PartialProxy:
    """
    A generic class that captures all the getattr (ie a.b) and getitem
    (ie a["b"]) calls on it
    """
    path: typing.List[typing.Callable] = field(default_factory=list)

    def _then(self, call: typing.Callable) -> "PartialProxy":
        """
        Return a copy of the current proxy, but with a new operation at the end
        :param call: A callable representing this last action
        """
        kwargs = {**self.__dict__, "path": self.path + [call]}
        return self.__class__(**kwargs)

    def eval(self, frame: dict) -> typing.Callable:
        """
        Evaluate the proxy calls on a given object and return the end of the chain
        :param frame: This is a dictionary of frame variables to evaluate on
        """
        current = frame

        # Whenever this is called, access each of the path elements one at a
        # time
        for segment in self.path:
            current = segment(current)

        return current

    def __getattr__(self, item):
        """
        Proxy an attribute get call
        """
        return self._then(operator.attrgetter(item))

    def __getitem__(self, item):
        """
        Proxy an item get call
        """
        return self._then(operator.itemgetter(item))



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['with_partial']

package_data = \
{'': ['*']}

install_requires = \
['addict>=2.4.0,<3.0.0']

extras_require = \
{'pipetools': ['pipetools>=1.0.1,<2.0.0']}

setup_kwargs = {
    'name': 'with-partial',
    'version': '0.1.0',
    'description': 'A utility for functional piping in Python that allows you to access any function in any scope as a partial.',
    'long_description': None,
    'author': 'Michael Milton',
    'author_email': 'michael.r.milton@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.7,<3.11',
}


setup(**setup_kwargs)
